@@ -2,7 +2,7 @@ package Inline::CPP::grammar;
 
 use strict;
 use vars qw($TYPEMAP_KIND $VERSION);
-$VERSION = '0.23';
+$VERSION = '0.24';
 
 #============================================================================
 # Regular expressions to match code blocks, numbers, strings, parenthesized
@@ -29,7 +29,7 @@ $string = qr'(?:(?:\")(?:[^\\\"]*(?:\\.[^\\\"]*)*)(?:\")|(?:\')(?:[^\\\']*(?:\\.
 
 # $RE{num}{real}|$RE{num}{real}{-base=>16}|$RE{num}{int}
 $number = qr'(?:(?i)(?:[+-]?)(?:(?=[0123456789]|[.])(?:[0123456789]*)(?:(?:[.])(?:[0123456789]{0,}))?)(?:(?:[E])(?:(?:[+-]?)(?:[0123456789]+))|))|(?:(?i)(?:[+-]?)(?:(?=[0123456789ABCDEF]|[.])(?:[0123456789ABCDEF]*)(?:(?:[.])(?:[0123456789ABCDEF]{0,}))?)(?:(?:[G])(?:(?:[+-]?)(?:[0123456789ABCDEF]+))|))|(?:(?:[+-]?)(?:\d+))';
-$funccall = qr/[_a-zA-Z][_a-zA-Z0-9]*(?:$Inline::CPP::grammar::parens)?/;
+$funccall = qr/(?:[_a-zA-Z][_a-zA-Z0-9]*::)*[_a-zA-Z][_a-zA-Z0-9]*(?:$Inline::CPP::grammar::parens)?/;
 
 #============================================================================
 # Inline::CPP's grammar
@@ -39,20 +39,59 @@ sub grammar {
 
 { use Data::Dumper; }
 
-code: part(s) {1}
-
-part: comment
-    | class_def
       {
-#         print "Found a class: $item[1]->[0]\n";
-         my $class = $item[1]->[0];
+    sub handle_class_def {
+         my ($thisparser, $def) = @_;
+#         print "Found a class: $def->[0]\n";
+         my $class = $def->[0];
          my @parts;
-         for my $part (@{$item[1]->[1]}) { push @parts, @$_ for @$part }
+         for my $part (@{$def->[1]}) { push @parts, @$_ for @$part }
          push @{$thisparser->{data}{classes}}, $class
            unless defined $thisparser->{data}{class}{$class};
          $thisparser->{data}{class}{$class} = \@parts;
-#	 print "Class:\n", Dumper \@parts;
+#	 print "Class $class:\n", Dumper \@parts;
 	 Inline::CPP::grammar::typemap($thisparser, $class);
+         [$class, \@parts];
+    }
+    sub handle_typedef {
+         my ($thisparser, $t) = @_;
+         my ($name, $type) = @{$t}{qw(name type)};
+#         print "found a typedef: $name => $type\n";
+
+         # XXX: this doesn't handle non-class typedefs that we could handle,
+         # e.g. "typedef int my_int_t"
+
+         if ($thisparser->{data}{class}{$type}
+             && !exists($thisparser->{data}{class}{$name})) {
+             push @{$thisparser->{data}{classes}}, $name;
+             $thisparser->{data}{class}{$name} = $thisparser->{data}{class}{$type};
+    	     Inline::CPP::grammar::typemap($thisparser, $name);
+         }
+         $t;
+    }
+    sub handle_enum {
+        my ($thisparser, $t) = @_;
+        $t;
+    }
+}
+
+code: part(s) {1}
+
+part: comment
+    | typedef
+      {
+        handle_typedef($thisparser, $item[1]);
+        1;
+      }
+    | enum
+      {
+        my $t = handle_enum($thisparser, $item[1]);
+        push @{$thisparser->{data}{enums}}, $t;
+        1;
+      }
+    | class_def
+      {
+         handle_class_def($thisparser, $item[1]);
 	 1;
       }
     | function_def
@@ -73,28 +112,53 @@ part: comment
       }
     | all
 
-class_def: class IDENTIFIER '{' class_part(s) '}' ';' 
+typedef: 'typedef' class IDENTIFIER(?) '{' <commit> class_part(s?) '}' IDENTIFIER ';'
+       {
+	 my ($class, $parts);
+         $class = $item[3][0] || 'anon_class'.($thisparser->{data}{anonclass}++);
+         ($class, $parts)= handle_class_def($thisparser, [$class, $item{class_part}]);
+	 { thing => 'typedef', name => $item[8], type => $class, body => $parts }
+       }
+       | 'typedef' IDENTIFIER IDENTIFIER ';'
+       { { thing => 'typedef', name => $item[3], type => $item[2] } }
+       | 'typedef' /[^;]*/ ';'
+       {
+#         dprint "Typedef $item{__DIRECTIVE1__} is too heinous\n";
+         { thing => 'comment'}
+       }
+
+enum: 'enum' IDENTIFIER(?) '{' <leftop: enum_item ',' enum_item> '}' ';'
+       {
+	{ thing => 'enum', name => $item{IDENTIFIER}[0],
+          body => $item{__DIRECTIVE1__} }
+       }
+
+enum_item: IDENTIFIER '=' <commit> /[0-9]+/
+         { [$item{IDENTIFIER}, $item{__PATTERN1__}] }
+         | IDENTIFIER
+         { [$item{IDENTIFIER}, undef] }
+
+class_def: class IDENTIFIER '{' <commit> class_part(s?) '}' ';' 
            { 
-#              print "Found a class definition: $item[2]\n";
- 	      [@item[2,4]]
+              [@item{'IDENTIFIER','class_part'}]
 	   }
-	 | class IDENTIFIER ':' <leftop: inherit ',' inherit> '{' class_part(s) '}' ';'
+	 | class IDENTIFIER ':' <commit> <leftop: inherit ',' inherit>
+            '{' class_part(s?) '}' ';'
 	   {
-#	       print "Found a class definition: $item[2]\n";
-	      push @{$item[6]}, [$item[4]]; 
-	      [@item[2,6]]
+	      push @{$item{class_part}}, [$item{__DIRECTIVE2__}];
+	      [@item{'IDENTIFIER','class_part'}]
 	   }
 
 inherit: scope IDENTIFIER 
 	{ {thing => 'inherits', name => $item[2], scope => $item[1]} }
 
 class_part: comment { [ {thing => 'comment'} ] }
-	  | scope ':' class_decl(s)
+	  | scope ':' <commit> class_decl(s?)
             {
-	      for my $part (@{$item[3]}) {
+	      for my $part (@{$item{class_decl}}) {
                   $_->{scope} = $item[1] for @$part;
 	      }
-	      $item[3]
+	      $item{class_decl}
 	    }
 	  | class_decl(s)
             {
@@ -106,6 +170,13 @@ class_part: comment { [ {thing => 'comment'} ] }
 	    }
 
 class_decl: comment { [{thing => 'comment'}] }
+          | typedef { [ handle_typedef($thisparser, $item[1]) ] }
+          | enum { [ handle_enum($thisparser, $item[1]) ] }
+          | class_def
+            {
+               my ($class, $parts) = handle_class_def($thisparser, $item[1]);
+               [{ thing => 'class', name => $class, body => $parts }];
+            }
           | method_def
 	    {
               $item[1]->{thing} = 'method';
@@ -125,19 +196,37 @@ class_decl: comment { [{thing => 'comment'}] }
 	      $item[1];
 	    }
 
-function_def: rtype IDENTIFIER '(' <leftop: arg ',' arg>(s?) ')' ';'
+function_def: operator <commit> ';'
               {
-                {rtype => $item[1], name => $item[2], args => $item[4]}
+                   $item[1]
               }
-            | rtype IDENTIFIER '(' <leftop: arg ',' arg>(s?) ')' code_block
+            | operator <commit> smod(?) code_block
               {
-                {rtype => $item[1], name => $item[2], args => $item[4]}
+                  $item[1]
+              }
+            | IDENTIFIER '(' <commit> <leftop: arg ',' arg>(s?) ')' smod(?) code_block
+              {
+                {name => $item{IDENTIFIER}, args => $item{__DIRECTIVE2__}, rtype => '' }
+              }
+            | rtype IDENTIFIER '(' <leftop: arg ',' arg>(s?) ')' ';'
+              {
+                {rtype => $item[1], name => $item[2], args => $item{__DIRECTIVE1__} }
+              }
+            | rtype IDENTIFIER '(' <leftop: arg ',' arg>(s?) ')' smod(?) code_block
+              {
+                {rtype => $item{rtype}, name => $item[2], args => $item{__DIRECTIVE1__} }
               }
 
-method_def: IDENTIFIER '(' <leftop: arg ',' arg>(s?) ')' method_imp
+method_def: operator <commit> method_imp
+            {
+#               print "method operator:\n", Dumper $item[1];
+               $item[1];
+            }
+
+          | IDENTIFIER '(' <commit> <leftop: arg ',' arg>(s?) ')' method_imp
             {
 #	      print "con-/de-structor found: $item[1]\n";
-              {name => $item[1], args => $item[3], abstract => ${$item[5]}};
+              {name => $item[1], args => $item{__DIRECTIVE2__}, abstract => ${$item{method_imp}} };
             }
           | rtype IDENTIFIER '(' <leftop: arg ',' arg>(s?) ')' method_imp
             {
@@ -150,6 +239,12 @@ method_def: IDENTIFIER '(' <leftop: arg ',' arg>(s?) ')' method_imp
 	      $thisparser->{data}{smod}{const} = 0; 
             }
 
+operator: rtype(?) 'operator' /\(\)|[^()]+/ '(' <leftop: arg ',' arg>(s?) ')'
+          {
+#            print "Found operator: $item[1][0] operator $item[3]\n";
+            {name=> "operator $item[3]", args => $item[5], ret => $item[1][0]}
+          }
+
 # By adding smod, we allow 'const' member functions. This would also bind to
 # incorrect C++ with the word 'static' after the argument list, but we don't
 # care at all because such code would never be compiled successfully.
@@ -158,8 +253,8 @@ method_def: IDENTIFIER '(' <leftop: arg ',' arg>(s?) ')' method_imp
 # allow them anywhere, but our goal is not to enforce c++ standards -- that's
 # the compiler's job.
 method_imp: smod(?) ';' { \0 }
+          | smod(?) '=' <commit> '0' ';' { \1 }
           | smod(?) initlist(?) code_block { \0 }
-          | smod(?) '=' '0' ';' { \1 }
           | smod(?) '=' '0' code_block { \0 }
 
 initlist: ':' <leftop: subexpr ',' subexpr>
@@ -181,25 +276,36 @@ var: star(s?) IDENTIFIER '=' expr { [@item[1,2]] }
 
 arg: type IDENTIFIER '=' expr
      { 
-#       print "argument $item[2] found\n";
-#       print "expression: $item[4]\n";
-	{type => $item[1], name => $item[2], optional => 1, 
+#       print "argument $item{IDENTIFIER} found\n";
+#       print "expression: $item{expr}\n";
+	{type => $item[1], name => $item{IDENTIFIER}, optional => 1, 
 	 offset => $thisoffset} 
      }
    | type IDENTIFIER
      { 
-#       print "argument $item[2] found\n";
-       {type => $item[1], name => $item[2], offset => $thisoffset}
+#       print "argument $item{IDENTIFIER} found\n";
+       {type => $item[1], name => $item{IDENTIFIER}, offset => $thisoffset}
      }
    | type { {type => $item[1]} }
    | '...' 
      { {name => '...', type => '...', offset => $thisoffset} }
 
-IDENTIFIER: /[~_a-z]\w*/i
-	    {
-#	      print "IDENTIFIER: $item[1]\n";
-	      $item[1]
-	    }
+ident_part: /[~_a-z]\w*/i '<' <commit> <leftop: IDENTIFIER ',' IDENTIFIER>(s?) '>'
+        {
+	   $item[1].'<'.join('', @{$item[4]}).'>'
+        }
+
+      | /[~_a-z]\w*/i
+        {
+           $item[1]
+        }
+
+IDENTIFIER: <leftop: ident_part '::' ident_part>
+ 	    {
+              my $x = join '::', @{$item[1]};
+#              print "IDENTIFIER: $x\n";
+              $x
+ 	    }
 
 # Parse::RecDescent is retarded in this one case: if a subrule fails, it
 # gives up the entire rule. This is a stupid way to get around that. 
@@ -209,8 +315,8 @@ rtype1: TYPE star(s?)
          $return = $item[1];
          $return .= join '',' ',@{$item[2]} if @{$item[2]};
 #	 print "rtype1: $return\n";
-         return undef 
-           unless(defined$thisparser->{data}{typeconv}{valid_rtypes}{$return});
+#          return undef 
+#            unless(defined$thisparser->{data}{typeconv}{valid_rtypes}{$return});
         }
 rtype2: modifier(s) TYPE star(s?)
 	{
@@ -219,8 +325,8 @@ rtype2: modifier(s) TYPE star(s?)
            if @{$item[1]};
          $return .= join '',' ',@{$item[3]} if @{$item[3]};
 #	 print "rtype2: $return\n";
-         return undef
-           unless(defined$thisparser->{data}{typeconv}{valid_rtypes}{$return});
+#          return undef
+#            unless(defined$thisparser->{data}{typeconv}{valid_rtypes}{$return});
 	 $return = 'static ' . $return
 	   if $thisparser->{data}{smod}{static};
          $thisparser->{data}{smod}{static} = 0;
@@ -230,17 +336,19 @@ type: type2 | type1
 type1: TYPE star(s?)
         {
          $return = $item[1];
-         $return .= join '',' ',@{$item[2]} if @{$item[2]};
-         return undef
-           unless(defined$thisparser->{data}{typeconv}{valid_types}{$return});
+         $return .= join '',' ',@{$item{star}} if @{$item{star}};
+#	 print "type1: $return\n";
+#          return undef
+#            unless(defined$thisparser->{data}{typeconv}{valid_types}{$return});
         }
 type2: modifier(s) TYPE star(s?)
 	{
-         $return = $item[2];
+         $return = $item{TYPE};
          $return = join ' ',grep{$_}@{$item[1]},$return if @{$item[1]};
-         $return .= join '',' ',@{$item[3]} if @{$item[3]};
-         return undef
-           unless(defined$thisparser->{data}{typeconv}{valid_types}{$return});
+         $return .= join '',' ',@{$item{star}} if @{$item{star}};
+#	 print "type2: $return\n";
+#          return undef
+#            unless(defined$thisparser->{data}{typeconv}{valid_types}{$return});
 	}
 
 anytype: anytype2 | anytype1
@@ -293,7 +401,7 @@ subexpr: /$Inline::CPP::grammar::funccall/ # Matches a macro, too
 OP: '+' | '-' | '*' | '/' | '^' | '&' | '|' | '%' | '||' | '&&'
 UOP: '~' | '!' | '-' | '*' | '&'
 
-TYPE: /\w+/
+TYPE: IDENTIFIER
 
 all: /.*/
 
